from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Tuple
import uuid

from common import OPENAPI_NO_VALIDATION, TestLogger
import pytest

from anyscale import Anyscale
from anyscale._private.anyscale_client import FakeAnyscaleClient
from anyscale._private.sdk.timer import FakeTimer
from anyscale.client.openapi_client.models import (
    Cloud,
    CloudProviders,
    ComputeNodeType as InternalApiComputeNodeType,
    ComputeTemplateConfig,
    DecoratedComputeTemplate,
    Resources,
    WorkerNodeType as InternalApiWorkerNodeType,
)
from anyscale.compute_config import ComputeConfigSDK
from anyscale.compute_config.models import (
    CloudDeployment,
    ComputeConfig,
    ComputeConfigVersion,
    HeadNodeConfig,
    MarketType,
    WorkerNodeGroupConfig,
)
from anyscale.sdk.anyscale_client.models import (
    ClusterCompute,
    ClusterComputeConfig,
    ComputeNodeType,
)


@pytest.fixture()
def compute_config_sdk_with_fakes(
    sdk_with_fakes: Tuple[Anyscale, FakeAnyscaleClient, TestLogger, FakeTimer]
) -> Tuple[ComputeConfigSDK, FakeAnyscaleClient]:
    sdk, client, _, _ = sdk_with_fakes
    return sdk.compute_config, client


class TestCreateComputeConfig:
    def test_name(
        self,
        compute_config_sdk_with_fakes: Tuple[ComputeConfigSDK, FakeAnyscaleClient],
    ):
        sdk, fake_client = compute_config_sdk_with_fakes

        # First version.
        full_name_1 = sdk.create(ComputeConfig(), name="test-compute-config-name")
        compute_config_id_1 = sdk.get(full_name_1).id
        assert full_name_1 == "test-compute-config-name:1"

        created_compute_config_1 = fake_client.get_compute_config(compute_config_id_1)
        assert created_compute_config_1 is not None
        assert created_compute_config_1.name == "test-compute-config-name"

        # Second version.
        full_name_2 = sdk.create(ComputeConfig(), name="test-compute-config-name")
        compute_config_id_2 = sdk.get(full_name_2).id
        assert full_name_2 == "test-compute-config-name:2"

        created_compute_config_2 = fake_client.get_compute_config(compute_config_id_2)
        assert created_compute_config_2 is not None
        assert created_compute_config_2.name == "test-compute-config-name"

    def test_name_with_version_tag(
        self,
        compute_config_sdk_with_fakes: Tuple[ComputeConfigSDK, FakeAnyscaleClient],
    ):
        sdk, fake_client = compute_config_sdk_with_fakes

        with pytest.raises(
            ValueError,
            match="A version tag cannot be provided when creating a compute config. The latest version tag will be generated and returned.",
        ):
            sdk.create(ComputeConfig(), name="test-compute-config-name:1")

    @pytest.mark.parametrize("use_custom_cloud", [False, True])
    @pytest.mark.parametrize("has_no_worker_nodes", [False, True])
    def test_no_head_node_uses_cloud_default(
        self,
        compute_config_sdk_with_fakes: Tuple[ComputeConfigSDK, FakeAnyscaleClient],
        use_custom_cloud: bool,
        has_no_worker_nodes: bool,
    ):
        sdk, fake_client = compute_config_sdk_with_fakes

        custom_cloud_name = "test-non-default-cloud"
        fake_client.add_cloud(
            Cloud(
                id=str(uuid.uuid4()),
                name=custom_cloud_name,
                local_vars_configuration=OPENAPI_NO_VALIDATION,
            ),
        )
        custom_cloud_id = fake_client.get_cloud_id(cloud_name=custom_cloud_name)
        fake_client.set_default_compute_config(
            ClusterCompute(
                id="test-custom-compute-config-id",
                config=ClusterComputeConfig(
                    cloud_id=custom_cloud_id,
                    head_node_type=ComputeNodeType(
                        name="non-default-head",
                        instance_type="custom-instance-type",
                        resources={"CPU": 24, "GPU": 2, "custom": 1},
                    ),
                    local_vars_configuration=OPENAPI_NO_VALIDATION,
                ),
                local_vars_configuration=OPENAPI_NO_VALIDATION,
            ),
            cloud_id=custom_cloud_id,
        )

        config = ComputeConfig()
        if has_no_worker_nodes:
            # Explicitly set no worker nodes.
            # Only in this case should the head node be schedulable.
            config = config.options(worker_nodes=[])
        if use_custom_cloud:
            config = config.options(cloud=custom_cloud_name)

        full_name = sdk.create(config, name="test123")
        compute_config_id = sdk.get(full_name).id
        created_compute_config = fake_client.get_compute_config(compute_config_id)
        assert created_compute_config is not None

        created = created_compute_config.config
        # Serverless worker config should only be set if worker_nodes is `None`.
        assert created.auto_select_worker_config is not has_no_worker_nodes

        if use_custom_cloud:
            assert created.cloud_id == custom_cloud_id
            assert created.head_node_type.instance_type == "custom-instance-type"
        else:
            assert created.cloud_id == fake_client.DEFAULT_CLOUD_ID
            default_compute_config = fake_client.get_default_compute_config(
                cloud_id=fake_client.DEFAULT_CLOUD_ID
            )
            assert (
                created.head_node_type.instance_type
                == default_compute_config.config.head_node_type.instance_type
            )

        if has_no_worker_nodes:
            assert created.head_node_type.resources is None
        else:
            assert created.head_node_type.resources == Resources(cpu=0, gpu=0)

    @pytest.mark.parametrize("use_custom_cloud", [False, True])
    @pytest.mark.parametrize("has_no_worker_nodes", [False, True])
    @pytest.mark.parametrize("has_resources", [False, True])
    def test_custom_head_node(
        self,
        compute_config_sdk_with_fakes: Tuple[ComputeConfigSDK, FakeAnyscaleClient],
        use_custom_cloud: bool,
        has_no_worker_nodes: bool,
        has_resources: bool,
    ):
        sdk, fake_client = compute_config_sdk_with_fakes

        custom_cloud_name = "test-non-default-cloud"
        fake_client.add_cloud(
            Cloud(
                id=str(uuid.uuid4()),
                name=custom_cloud_name,
                local_vars_configuration=OPENAPI_NO_VALIDATION,
            ),
        )
        custom_cloud_id = fake_client.get_cloud_id(cloud_name=custom_cloud_name)

        head_node_config = HeadNodeConfig(instance_type="head-node-instance-type",)
        if has_resources:
            head_node_config = head_node_config.options(
                resources={"CPU": 1, "head_node": 1}
            )

        config = ComputeConfig(head_node=head_node_config)
        if has_no_worker_nodes:
            # Explicitly set no worker nodes.
            # Only in this case should the head node be schedulable.
            config = config.options(worker_nodes=[])
        if use_custom_cloud:
            config = config.options(cloud=custom_cloud_name)

        full_name = sdk.create(config, name="test123")
        compute_config_id = sdk.get(full_name).id
        created_compute_config = fake_client.get_compute_config(compute_config_id)
        assert created_compute_config is not None

        created = created_compute_config.config
        if use_custom_cloud:
            assert created.cloud_id == custom_cloud_id
        else:
            assert created.cloud_id == fake_client.DEFAULT_CLOUD_ID

        assert created.head_node_type.instance_type == "head-node-instance-type"

        # Serverless worker config should only be set if worker_nodes is `None`.
        assert created.auto_select_worker_config is not has_no_worker_nodes

        # If the user explicitly provides resources, they should always be set.
        if has_resources:
            assert created.head_node_type.resources == Resources(
                cpu=1, custom_resources={"head_node": 1}
            )
        # If there are no worker nodes, resources should be empty (populated by backend).
        elif has_no_worker_nodes:
            assert created.head_node_type.resources is None
        # Otherwise, head node is unschedulable by default.
        else:
            assert created.head_node_type.resources == Resources(cpu=0, gpu=0)

    @pytest.mark.parametrize(
        "provider",
        [CloudProviders.AWS, CloudProviders.GCP, CloudProviders.CLOUDGATEWAY],
    )
    @pytest.mark.parametrize("advanced_instance_config", [None, {}, {"foo": "bar"}])
    @pytest.mark.parametrize(
        "location", ["TOP_LEVEL", "HEAD", "WORKER"],
    )
    def test_advanced_instance_config(  # noqa: PLR0912
        self,
        compute_config_sdk_with_fakes: Tuple[ComputeConfigSDK, FakeAnyscaleClient],
        provider: CloudProviders,
        advanced_instance_config: Optional[Dict],
        location: str,
    ):
        sdk, fake_client = compute_config_sdk_with_fakes

        custom_cloud_name = "test-non-default-cloud"
        fake_client.add_cloud(
            Cloud(
                id=str(uuid.uuid4()),
                name=custom_cloud_name,
                provider=provider,
                local_vars_configuration=OPENAPI_NO_VALIDATION,
            ),
        )
        custom_cloud_id = fake_client.get_cloud_id(cloud_name=custom_cloud_name)

        config = ComputeConfig(
            cloud=custom_cloud_name,
            advanced_instance_config=advanced_instance_config
            if location == "TOP_LEVEL"
            else None,
            head_node=HeadNodeConfig(
                instance_type="head-node-instance-type",
                advanced_instance_config=advanced_instance_config
                if location == "HEAD"
                else None,
            ),
            worker_nodes=[
                WorkerNodeGroupConfig(
                    instance_type="worker-node-instance-type",
                    advanced_instance_config=advanced_instance_config
                    if location == "WORKER"
                    else None,
                ),
            ],
        )

        if advanced_instance_config and provider not in {
            CloudProviders.AWS,
            CloudProviders.GCP,
        }:
            with pytest.raises(
                ValueError,
                match=f"'advanced_instance_config' is not implemented for cloud provider: '{provider}'",
            ):
                sdk.create(config, name="test123")

            return

        full_name = sdk.create(config, name="test123")
        compute_config_id = sdk.get(full_name).id
        created_compute_config = fake_client.get_compute_config(compute_config_id)
        assert created_compute_config is not None

        created = created_compute_config.config
        assert created.cloud_id == custom_cloud_id
        assert created.head_node_type.instance_type == "head-node-instance-type"
        if not advanced_instance_config:
            assert created.aws_advanced_configurations_json is None
            assert created.gcp_advanced_configurations_json is None
            assert created.head_node_type.aws_advanced_configurations_json is None
            assert created.head_node_type.gcp_advanced_configurations_json is None
            assert created.worker_node_types[0].aws_advanced_configurations_json is None
            assert created.worker_node_types[0].gcp_advanced_configurations_json is None
        elif provider == CloudProviders.AWS:
            if location == "TOP_LEVEL":
                assert (
                    created.aws_advanced_configurations_json == advanced_instance_config
                )
            else:
                assert created.aws_advanced_configurations_json is None
            assert created.gcp_advanced_configurations_json is None

            if location == "HEAD":
                assert (
                    created.head_node_type.aws_advanced_configurations_json
                    == advanced_instance_config
                )
            else:
                assert created.head_node_type.aws_advanced_configurations_json is None
            assert created.head_node_type.gcp_advanced_configurations_json is None

            if location == "WORKER":
                assert (
                    created.worker_node_types[0].aws_advanced_configurations_json
                    == advanced_instance_config
                )
            else:
                assert (
                    created.worker_node_types[0].aws_advanced_configurations_json
                    is None
                )
            assert created.worker_node_types[0].gcp_advanced_configurations_json is None

        elif provider == CloudProviders.GCP:
            assert created.aws_advanced_configurations_json is None
            if location == "TOP_LEVEL":
                assert (
                    created.gcp_advanced_configurations_json == advanced_instance_config
                )
            else:
                assert created.gcp_advanced_configurations_json is None

            assert created.head_node_type.aws_advanced_configurations_json is None
            if location == "HEAD":
                assert (
                    created.head_node_type.gcp_advanced_configurations_json
                    == advanced_instance_config
                )
            else:
                assert created.head_node_type.gcp_advanced_configurations_json is None

            assert created.worker_node_types[0].aws_advanced_configurations_json is None
            if location == "WORKER":
                assert (
                    created.worker_node_types[0].gcp_advanced_configurations_json
                    == advanced_instance_config
                )
            else:
                assert (
                    created.worker_node_types[0].gcp_advanced_configurations_json
                    is None
                )
        else:
            raise AssertionError(f"Unexpected provider '{provider}'")

    @pytest.mark.parametrize("use_custom_cloud", [False, True])
    @pytest.mark.parametrize("enable_cross_zone_scaling", [False, True])
    @pytest.mark.parametrize("zones", [None, ["zone1", "zone2"]])
    @pytest.mark.parametrize(
        "min_resources", [None, {"CPU": 1, "GPU": 1, "CUSTOM_RESOURCE": 1}]
    )
    @pytest.mark.parametrize(
        "max_resources", [None, {"CPU": 10, "GPU": 5, "CUSTOM_RESOURCE": 15}]
    )
    @pytest.mark.parametrize(
        "cluster_flags",
        [None, {"fake-cluster-feature-1": True, "fake-cluster-feature-2": "yes"}],
    )
    @pytest.mark.parametrize(
        "head_node_flags",
        [None, {"fake-head-node-feature-1": False, "fake-head-node-feature-2": "no"}],
    )
    @pytest.mark.parametrize(
        "head_node_cloud_deployment",
        [
            None,
            {"provider": "fake-provider", "machine_pool": "fake-machine-pool",},
            {
                "provider": "fake-provider",
                "region": "fake-region",
                "machine_pool": "fake-machine-pool",
                "id": "fake-id",
            },
        ],
    )
    def test_flags(  # noqa: PLR0912 PLR0913
        self,
        compute_config_sdk_with_fakes: Tuple[ComputeConfigSDK, FakeAnyscaleClient],
        use_custom_cloud: bool,
        enable_cross_zone_scaling: bool,
        zones: Optional[List[str]],
        min_resources: Optional[Dict[str, float]],
        max_resources: Optional[Dict[str, float]],
        cluster_flags: Optional[Dict[str, Any]],
        head_node_flags: Optional[Dict[str, Any]],
        head_node_cloud_deployment: Optional[CloudDeployment],
    ):
        sdk, fake_client = compute_config_sdk_with_fakes

        custom_cloud_name = "test-non-default-cloud"
        fake_client.add_cloud(
            Cloud(
                id=str(uuid.uuid4()),
                name=custom_cloud_name,
                local_vars_configuration=OPENAPI_NO_VALIDATION,
            ),
        )
        custom_cloud_id = fake_client.get_cloud_id(cloud_name=custom_cloud_name)

        head_node_config = HeadNodeConfig(instance_type="head-node-instance-type",)
        if head_node_flags:
            head_node_config = head_node_config.options(flags=head_node_flags)
        if head_node_cloud_deployment:
            head_node_config = head_node_config.options(
                cloud_deployment=head_node_cloud_deployment
            )

        config = ComputeConfig(
            head_node=head_node_config,
            zones=zones,
            min_resources=min_resources,
            max_resources=max_resources,
        )
        if use_custom_cloud:
            config = config.options(cloud=custom_cloud_name)
        if enable_cross_zone_scaling:
            config = config.options(enable_cross_zone_scaling=True)
        if cluster_flags:
            config = config.options(flags=cluster_flags)

        full_name = sdk.create(config, name="test123")
        compute_config_id = sdk.get(full_name).id
        created_compute_config = fake_client.get_compute_config(compute_config_id)
        assert created_compute_config is not None

        created = created_compute_config.config
        if use_custom_cloud:
            assert created.cloud_id == custom_cloud_id
        else:
            assert created.cloud_id == fake_client.DEFAULT_CLOUD_ID

        assert created.head_node_type.instance_type == "head-node-instance-type"
        if head_node_flags:
            for key, value in head_node_flags.items():
                assert created.head_node_type.flags[key] == value
        if head_node_cloud_deployment:
            assert len(created.head_node_type.flags["cloud_deployment"]) == len(
                head_node_cloud_deployment
            )
            for key, value in head_node_cloud_deployment.items():
                assert created.head_node_type.flags["cloud_deployment"][key] == value

        assert (
            created.flags["allow-cross-zone-autoscaling"] == enable_cross_zone_scaling
        )
        assert created.allowed_azs == zones
        assert created.auto_select_worker_config is True
        if min_resources is None:
            assert "min_resources" not in created.flags
        else:
            assert len(created.flags["min_resources"]) == len(min_resources)
            for key, value in min_resources.items():
                assert created.flags["min_resources"][key] == value
        if max_resources is None:
            assert "max_resources" not in created.flags
        else:
            assert len(created.flags["max_resources"]) == len(max_resources)
            for key, value in max_resources.items():
                assert created.flags["max_resources"][key] == value
        if cluster_flags:
            for key, value in cluster_flags.items():
                assert created.flags[key] == value

    def test_custom_worker_nodes(
        self,
        compute_config_sdk_with_fakes: Tuple[ComputeConfigSDK, FakeAnyscaleClient],
    ):
        sdk, fake_client = compute_config_sdk_with_fakes
        config = ComputeConfig(
            worker_nodes=[
                WorkerNodeGroupConfig(instance_type="instance-type-1",),
                WorkerNodeGroupConfig(
                    name="group2",
                    instance_type="instance-type-2",
                    min_nodes=0,
                    max_nodes=100,
                    market_type=MarketType.SPOT,
                ),
                WorkerNodeGroupConfig(
                    name="group3",
                    instance_type="instance-type-2",
                    min_nodes=0,
                    max_nodes=100,
                    resources={"CPU": 1000, "custom": 1},
                    market_type=MarketType.PREFER_SPOT,
                    flags={
                        "fake-worker-node-feature-1": False,
                        "fake-worker-node-feature-2": "no",
                    },
                    cloud_deployment=CloudDeployment(
                        provider="fake-provider",
                        region="fake-region",
                        machine_pool="fake-machine-pool",
                        id="fake-id",
                    ),
                ),
            ],
        )

        full_name = sdk.create(config, name="test123")
        compute_config_id = sdk.get(full_name).id
        created_compute_config = fake_client.get_compute_config(compute_config_id)
        assert created_compute_config is not None

        created = created_compute_config.config

        # Serverless worker config should not be set if worker nodes are provided.
        assert created.auto_select_worker_config is False

        assert created.worker_node_types[0].name == "instance-type-1"
        assert created.worker_node_types[0].instance_type == "instance-type-1"
        assert created.worker_node_types[0].resources is None
        assert created.worker_node_types[0].min_workers == 0
        assert created.worker_node_types[0].max_workers == 10
        assert created.worker_node_types[0].use_spot is False
        assert created.worker_node_types[0].fallback_to_ondemand is False

        assert created.worker_node_types[1].name == "group2"
        assert created.worker_node_types[1].instance_type == "instance-type-2"
        assert created.worker_node_types[1].resources is None
        assert created.worker_node_types[1].min_workers == 0
        assert created.worker_node_types[1].max_workers == 100
        assert created.worker_node_types[1].use_spot is True
        assert created.worker_node_types[1].fallback_to_ondemand is False

        assert created.worker_node_types[2].name == "group3"
        assert created.worker_node_types[2].instance_type == "instance-type-2"
        assert created.worker_node_types[2].resources == Resources(
            cpu=1000, custom_resources={"custom": 1}
        )
        assert created.worker_node_types[2].min_workers == 0
        assert created.worker_node_types[2].max_workers == 100
        assert created.worker_node_types[2].use_spot is True
        assert created.worker_node_types[2].fallback_to_ondemand is True

        assert created.worker_node_types[2].flags["fake-worker-node-feature-1"] is False
        assert created.worker_node_types[2].flags["fake-worker-node-feature-2"] == "no"

        assert (
            created.worker_node_types[2].flags["cloud_deployment"]["provider"]
            == "fake-provider"
        )
        assert (
            created.worker_node_types[2].flags["cloud_deployment"]["region"]
            == "fake-region"
        )
        assert (
            created.worker_node_types[2].flags["cloud_deployment"]["machine_pool"]
            == "fake-machine-pool"
        )
        assert created.worker_node_types[2].flags["cloud_deployment"]["id"] == "fake-id"


@dataclass
class ResourcesTestCase:
    api_resources: Optional[Resources]
    expected_resources_dict: Optional[Dict[str, float]]


RESOURCES_TEST_CASES = [
    ResourcesTestCase(None, None),
    ResourcesTestCase(Resources(), {}),
    ResourcesTestCase(Resources(cpu=1), {"CPU": 1}),
    ResourcesTestCase(
        Resources(cpu=1, gpu=2, memory=1024, object_store_memory=1024 ** 2),
        {"CPU": 1, "GPU": 2, "memory": 1024, "object_store_memory": 1024 ** 2},
    ),
    # Keys with `None` values should be omitted.
    ResourcesTestCase(Resources(cpu=1, gpu=None), {"CPU": 1}),
    # custom_resources field should be flattened.
    ResourcesTestCase(
        Resources(cpu=1, custom_resources={"custom": 123}), {"CPU": 1, "custom": 123}
    ),
]


class TestGetComputeConfig:
    def test_no_name_or_id(
        self,
        compute_config_sdk_with_fakes: Tuple[ComputeConfigSDK, FakeAnyscaleClient],
    ):
        sdk, _ = compute_config_sdk_with_fakes
        with pytest.raises(ValueError, match="Either name or ID must be provided."):
            sdk.get(name="")

    def test_not_found(
        self,
        compute_config_sdk_with_fakes: Tuple[ComputeConfigSDK, FakeAnyscaleClient],
    ):
        sdk, fake_client = compute_config_sdk_with_fakes
        with pytest.raises(
            RuntimeError, match="Compute config 'does-not-exist' not found."
        ):
            sdk.get(name="does-not-exist")

        with pytest.raises(
            RuntimeError, match="Compute config with ID 'does-not-exist' not found."
        ):
            sdk.get(name="", _id="does-not-exist")

    @pytest.mark.parametrize("by_id", [False, True])
    def test_cloud_name(
        self,
        compute_config_sdk_with_fakes: Tuple[ComputeConfigSDK, FakeAnyscaleClient],
        by_id: bool,
    ):
        sdk, fake_client = compute_config_sdk_with_fakes
        fake_client.add_compute_config(
            DecoratedComputeTemplate(
                id="default-cloud-compute-config-id",
                name="default-cloud-compute-config-name",
                config=ComputeTemplateConfig(
                    cloud_id=fake_client.DEFAULT_CLOUD_ID,
                    head_node_type=InternalApiComputeNodeType(
                        name="head-node-name", instance_type="head-node-instance-type",
                    ),
                    worker_node_types=[],
                    local_vars_configuration=OPENAPI_NO_VALIDATION,
                ),
                local_vars_configuration=OPENAPI_NO_VALIDATION,
            )
        )

        default_cloud_config: ComputeConfig = sdk.get(
            _id="default-cloud-compute-config-id" if by_id else None,
            name="default-cloud-compute-config-name" if not by_id else "",
        ).config
        assert default_cloud_config.cloud == fake_client.DEFAULT_CLOUD_NAME

        fake_client.add_cloud(
            Cloud(
                id="fake-custom-cloud-id",
                name="fake-custom-cloud",
                local_vars_configuration=OPENAPI_NO_VALIDATION,
            ),
        )
        fake_client.add_compute_config(
            DecoratedComputeTemplate(
                id="custom-cloud-compute-config-id",
                name="custom-cloud-compute-config-name",
                config=ComputeTemplateConfig(
                    cloud_id="fake-custom-cloud-id",
                    head_node_type=InternalApiComputeNodeType(
                        name="head-node-name", instance_type="head-node-instance-type",
                    ),
                    worker_node_types=[],
                    local_vars_configuration=OPENAPI_NO_VALIDATION,
                ),
                local_vars_configuration=OPENAPI_NO_VALIDATION,
            )
        )

        custom_cloud_config: ComputeConfig = sdk.get(
            _id="custom-cloud-compute-config-id" if by_id else None,
            name="custom-cloud-compute-config-name" if not by_id else None,
        ).config
        assert custom_cloud_config.cloud == "fake-custom-cloud"

    @pytest.mark.parametrize(
        ("api_zones", "expected_zones"),
        [
            (None, None),
            ([], None),
            # API returns ["any"] if no zones are passed in.
            (["any"], None),
            (["az1"], ["az1"]),
            (["az1", "az2"], ["az1", "az2"]),
        ],
    )
    def test_zones(
        self,
        compute_config_sdk_with_fakes: Tuple[ComputeConfigSDK, FakeAnyscaleClient],
        api_zones: Optional[List[str]],
        expected_zones: Optional[List[str]],
    ):
        sdk, fake_client = compute_config_sdk_with_fakes
        fake_client.add_compute_config(
            DecoratedComputeTemplate(
                id="fake-compute-config-id",
                name="fake-compute-config-name",
                config=ComputeTemplateConfig(
                    cloud_id=fake_client.DEFAULT_CLOUD_ID,
                    allowed_azs=api_zones,
                    head_node_type=InternalApiComputeNodeType(
                        name="head-node-name", instance_type="head-node-instance-type",
                    ),
                    worker_node_types=[],
                    local_vars_configuration=OPENAPI_NO_VALIDATION,
                ),
                local_vars_configuration=OPENAPI_NO_VALIDATION,
            )
        )

        config: ComputeConfig = sdk.get(name="fake-compute-config-name").config
        assert config.zones == expected_zones

    @pytest.mark.parametrize(
        ("flags", "expected"),
        [
            (None, False),
            ({}, False),
            ({"something-else": "foobar"}, False),
            ({"allow-cross-zone-autoscaling": False}, False),
            ({"allow-cross-zone-autoscaling": True}, True),
        ],
    )
    def test_enable_cross_zone_scaling(
        self,
        compute_config_sdk_with_fakes: Tuple[ComputeConfigSDK, FakeAnyscaleClient],
        flags: Optional[Dict],
        expected: bool,
    ):
        sdk, fake_client = compute_config_sdk_with_fakes
        fake_client.add_compute_config(
            DecoratedComputeTemplate(
                id="fake-compute-config-id",
                name="fake-compute-config-name",
                config=ComputeTemplateConfig(
                    cloud_id=fake_client.DEFAULT_CLOUD_ID,
                    flags=flags,
                    head_node_type=InternalApiComputeNodeType(
                        name="head-node-name", instance_type="head-node-instance-type",
                    ),
                    worker_node_types=[],
                    local_vars_configuration=OPENAPI_NO_VALIDATION,
                ),
                local_vars_configuration=OPENAPI_NO_VALIDATION,
            )
        )

        config: ComputeConfig = sdk.get(name="fake-compute-config-name").config
        assert config.enable_cross_zone_scaling == expected

    @pytest.mark.parametrize(
        ("flags", "expected"),
        [
            (None, None),
            ({}, None),
            ({"max-cpus": None, "max-gpus": None}, None),
            ({"max-cpus": 10, "max-gpus": None}, {"CPU": 10}),
            ({"max-cpus": None, "max-gpus": 5}, {"GPU": 5}),
            ({"max-cpus": 10, "max-gpus": 5}, {"CPU": 10, "GPU": 5}),
        ],
    )
    def test_max_resources(
        self,
        compute_config_sdk_with_fakes: Tuple[ComputeConfigSDK, FakeAnyscaleClient],
        flags: Optional[Dict],
        expected: Optional[Dict],
    ):
        sdk, fake_client = compute_config_sdk_with_fakes
        fake_client.add_compute_config(
            DecoratedComputeTemplate(
                id="fake-compute-config-id",
                name="fake-compute-config-name",
                config=ComputeTemplateConfig(
                    cloud_id=fake_client.DEFAULT_CLOUD_ID,
                    flags=flags,
                    head_node_type=InternalApiComputeNodeType(
                        name="head-node-name", instance_type="head-node-instance-type",
                    ),
                    worker_node_types=[],
                    local_vars_configuration=OPENAPI_NO_VALIDATION,
                ),
                local_vars_configuration=OPENAPI_NO_VALIDATION,
            )
        )

        config: ComputeConfig = sdk.get(name="fake-compute-config-name").config
        assert config.max_resources == expected

    def test_auto_select_worker_config(
        self,
        compute_config_sdk_with_fakes: Tuple[ComputeConfigSDK, FakeAnyscaleClient],
    ):
        sdk, fake_client = compute_config_sdk_with_fakes

        fake_client.add_compute_config(
            DecoratedComputeTemplate(
                id="serverless-compute-config-id",
                name="serverless-compute-config-name",
                config=ComputeTemplateConfig(
                    cloud_id=fake_client.DEFAULT_CLOUD_ID,
                    head_node_type=InternalApiComputeNodeType(
                        name="head-node-name", instance_type="head-node-instance-type",
                    ),
                    worker_node_types=[],
                    auto_select_worker_config=True,
                    local_vars_configuration=OPENAPI_NO_VALIDATION,
                ),
                local_vars_configuration=OPENAPI_NO_VALIDATION,
            )
        )

        serverless_config: ComputeConfig = sdk.get(
            name="serverless-compute-config-name"
        ).config
        assert serverless_config.worker_nodes is None

        fake_client.add_compute_config(
            DecoratedComputeTemplate(
                id="non-serverless-compute-config-id",
                name="non-serverless-compute-config-name",
                config=ComputeTemplateConfig(
                    cloud_id=fake_client.DEFAULT_CLOUD_ID,
                    head_node_type=InternalApiComputeNodeType(
                        name="head-node-name", instance_type="head-node-instance-type",
                    ),
                    worker_node_types=[],
                    auto_select_worker_config=False,
                    local_vars_configuration=OPENAPI_NO_VALIDATION,
                ),
                local_vars_configuration=OPENAPI_NO_VALIDATION,
            )
        )

        non_serverless_config: ComputeConfig = sdk.get(
            name="non-serverless-compute-config-name"
        ).config
        assert non_serverless_config.worker_nodes == []

    @pytest.mark.parametrize("test_case", RESOURCES_TEST_CASES)
    def test_convert_head_node(
        self,
        compute_config_sdk_with_fakes: Tuple[ComputeConfigSDK, FakeAnyscaleClient],
        test_case: ResourcesTestCase,
    ):
        sdk, fake_client = compute_config_sdk_with_fakes

        fake_client.add_compute_config(
            DecoratedComputeTemplate(
                id="fake-compute-config-id",
                name="fake-compute-config-name",
                config=ComputeTemplateConfig(
                    cloud_id=fake_client.DEFAULT_CLOUD_ID,
                    head_node_type=InternalApiComputeNodeType(
                        name="head-node-name",
                        instance_type="head-node-instance-type",
                        resources=test_case.api_resources,
                    ),
                    worker_node_types=[],
                    local_vars_configuration=OPENAPI_NO_VALIDATION,
                ),
                local_vars_configuration=OPENAPI_NO_VALIDATION,
            )
        )

        config: ComputeConfig = sdk.get(name="fake-compute-config-name").config
        assert config.head_node == HeadNodeConfig(
            instance_type="head-node-instance-type",
            resources=test_case.expected_resources_dict,
        )

    @pytest.mark.parametrize("test_case", RESOURCES_TEST_CASES)
    def test_convert_worker_nodes(
        self,
        compute_config_sdk_with_fakes: Tuple[ComputeConfigSDK, FakeAnyscaleClient],
        test_case: ResourcesTestCase,
    ):
        sdk, fake_client = compute_config_sdk_with_fakes

        fake_client.add_compute_config(
            DecoratedComputeTemplate(
                id="fake-compute-config-id",
                name="fake-compute-config-name",
                config=ComputeTemplateConfig(
                    cloud_id=fake_client.DEFAULT_CLOUD_ID,
                    head_node_type=InternalApiComputeNodeType(
                        name="head-node-name", instance_type="head-node-instance-type",
                    ),
                    worker_node_types=[
                        InternalApiWorkerNodeType(
                            name="basic",
                            instance_type="instance-type-1",
                            min_workers=0,
                            max_workers=10,
                        ),
                        InternalApiWorkerNodeType(
                            name="custom-resources",
                            instance_type="instance-type-2",
                            resources=test_case.api_resources,
                            min_workers=1,
                            max_workers=1,
                        ),
                        InternalApiWorkerNodeType(
                            name="min-workers-none",
                            instance_type="instance-type-3",
                            min_workers=None,
                            max_workers=1,
                        ),
                        InternalApiWorkerNodeType(
                            name="max-workers-none",
                            instance_type="instance-type-4",
                            min_workers=0,
                            max_workers=None,
                        ),
                    ],
                    local_vars_configuration=OPENAPI_NO_VALIDATION,
                ),
                local_vars_configuration=OPENAPI_NO_VALIDATION,
            )
        )

        config: ComputeConfig = sdk.get(name="fake-compute-config-name").config
        assert config.worker_nodes == [
            WorkerNodeGroupConfig(
                name="basic",
                instance_type="instance-type-1",
                min_nodes=0,
                max_nodes=10,
            ),
            WorkerNodeGroupConfig(
                name="custom-resources",
                instance_type="instance-type-2",
                resources=test_case.expected_resources_dict,
                min_nodes=1,
                max_nodes=1,
            ),
            WorkerNodeGroupConfig(
                name="min-workers-none",
                instance_type="instance-type-3",
                min_nodes=0,
                max_nodes=1,
            ),
            WorkerNodeGroupConfig(
                name="max-workers-none",
                instance_type="instance-type-4",
                min_nodes=0,
                max_nodes=10,
            ),
        ]

    def test_worker_node_market_type(
        self,
        compute_config_sdk_with_fakes: Tuple[ComputeConfigSDK, FakeAnyscaleClient],
    ):
        sdk, fake_client = compute_config_sdk_with_fakes

        fake_client.add_compute_config(
            DecoratedComputeTemplate(
                id="fake-compute-config-id",
                name="fake-compute-config-name",
                config=ComputeTemplateConfig(
                    cloud_id=fake_client.DEFAULT_CLOUD_ID,
                    head_node_type=InternalApiComputeNodeType(
                        name="head-node-name", instance_type="head-node-instance-type",
                    ),
                    worker_node_types=[
                        InternalApiWorkerNodeType(
                            name="on-demand-worker-node-group",
                            instance_type="on-demand-worker-node-group",
                            min_workers=1,
                            max_workers=1,
                            use_spot=False,
                            fallback_to_ondemand=False,
                        ),
                        InternalApiWorkerNodeType(
                            name="spot-worker-node-group",
                            instance_type="spot-worker-node-group",
                            min_workers=1,
                            max_workers=1,
                            use_spot=True,
                            fallback_to_ondemand=False,
                        ),
                        InternalApiWorkerNodeType(
                            name="prefer-spot-worker-node-group",
                            instance_type="prefer-spot-worker-node-group",
                            min_workers=1,
                            max_workers=1,
                            use_spot=True,
                            fallback_to_ondemand=True,
                        ),
                    ],
                    local_vars_configuration=OPENAPI_NO_VALIDATION,
                ),
                local_vars_configuration=OPENAPI_NO_VALIDATION,
            )
        )

        config: ComputeConfig = sdk.get(name="fake-compute-config-name").config
        assert config.worker_nodes == [
            WorkerNodeGroupConfig(
                name="on-demand-worker-node-group",
                instance_type="on-demand-worker-node-group",
                min_nodes=1,
                max_nodes=1,
                market_type=MarketType.ON_DEMAND,
            ),
            WorkerNodeGroupConfig(
                name="spot-worker-node-group",
                instance_type="spot-worker-node-group",
                min_nodes=1,
                max_nodes=1,
                market_type=MarketType.SPOT,
            ),
            WorkerNodeGroupConfig(
                name="prefer-spot-worker-node-group",
                instance_type="prefer-spot-worker-node-group",
                min_nodes=1,
                max_nodes=1,
                market_type=MarketType.PREFER_SPOT,
            ),
        ]

    @pytest.mark.parametrize(
        "provider", [CloudProviders.AWS, CloudProviders.GCP],
    )
    @pytest.mark.parametrize("advanced_instance_config", [None, {}, {"foo": "bar"}])
    @pytest.mark.parametrize(
        "location", ["TOP_LEVEL", "HEAD", "WORKER"],
    )
    def test_advanced_instance_config(
        self,
        compute_config_sdk_with_fakes: Tuple[ComputeConfigSDK, FakeAnyscaleClient],
        provider: CloudProviders,
        advanced_instance_config: Optional[Dict],
        location: str,
    ):
        sdk, fake_client = compute_config_sdk_with_fakes
        fake_client.add_cloud(
            Cloud(
                id="fake-custom-cloud-id",
                name="fake-custom-cloud",
                provider=provider,
                local_vars_configuration=OPENAPI_NO_VALIDATION,
            ),
        )

        aws_config = (
            advanced_instance_config if provider == CloudProviders.AWS else None
        )
        gcp_config = (
            advanced_instance_config if provider == CloudProviders.GCP else None
        )
        fake_client.add_compute_config(
            DecoratedComputeTemplate(
                id="fake-compute-config-id",
                name="fake-compute-config-name",
                config=ComputeTemplateConfig(
                    cloud_id="fake-custom-cloud-id",
                    aws_advanced_configurations_json=aws_config
                    if location == "TOP_LEVEL"
                    else None,
                    gcp_advanced_configurations_json=gcp_config
                    if location == "TOP_LEVEL"
                    else None,
                    head_node_type=InternalApiComputeNodeType(
                        name="head-node-name",
                        instance_type="head-node-instance-type",
                        aws_advanced_configurations_json=aws_config
                        if location == "HEAD"
                        else None,
                        gcp_advanced_configurations_json=gcp_config
                        if location == "HEAD"
                        else None,
                    ),
                    worker_node_types=[
                        InternalApiWorkerNodeType(
                            name="worker-node-group",
                            instance_type="worker-node-group",
                            aws_advanced_configurations_json=aws_config
                            if location == "WORKER"
                            else None,
                            gcp_advanced_configurations_json=gcp_config
                            if location == "WORKER"
                            else None,
                        ),
                    ],
                    local_vars_configuration=OPENAPI_NO_VALIDATION,
                ),
                local_vars_configuration=OPENAPI_NO_VALIDATION,
            )
        )

        config: ComputeConfig = sdk.get(name="fake-compute-config-name").config
        assert config.advanced_instance_config == (
            advanced_instance_config or None if location == "TOP_LEVEL" else None
        )
        assert isinstance(config.head_node, HeadNodeConfig)
        assert config.head_node.advanced_instance_config == (
            advanced_instance_config or None if location == "HEAD" else None
        )
        assert isinstance(config.worker_nodes, list) and isinstance(
            config.worker_nodes[0], WorkerNodeGroupConfig
        )
        assert config.worker_nodes[0].advanced_instance_config == (
            advanced_instance_config or None if location == "WORKER" else None
        )


class TestArchive:
    def test_not_found(
        self, compute_config_sdk_with_fakes: Tuple[ComputeConfigSDK, FakeAnyscaleClient]
    ):
        sdk, _ = compute_config_sdk_with_fakes

        with pytest.raises(
            RuntimeError, match="Compute config 'does-not-exist' not found"
        ):
            sdk.archive(name="does-not-exist")

    @pytest.mark.parametrize("use_full_name", [False, True])
    def test_basic(
        self,
        compute_config_sdk_with_fakes: Tuple[ComputeConfigSDK, FakeAnyscaleClient],
        use_full_name: bool,
    ):
        sdk, fake_client = compute_config_sdk_with_fakes

        full_name = sdk.create(ComputeConfig(), name="test-compute-config-name")
        compute_config_id = sdk.get(full_name).id

        assert full_name == "test-compute-config-name:1"

        assert not fake_client.is_archived_compute_config(compute_config_id)

        if use_full_name:
            sdk.archive(name=full_name)
        else:
            sdk.archive(name="test-compute-config-name")

        assert fake_client.is_archived_compute_config(compute_config_id)

        with pytest.raises(
            RuntimeError, match="Compute config 'test-compute-config-name' not found"
        ):
            sdk.get(name="test-compute-config-name")

        archived_version: ComputeConfigVersion = sdk.get(
            name="test-compute-config-name", include_archived=True
        )
        assert archived_version.config is not None



import os
import re
import tempfile
from typing import Optional, Tuple
from unittest import mock
from unittest.mock import patch

from common import TEST_CONTAINERFILE, TestLogger
import pytest
from tests.unit.common import verify_file_content

from anyscale import Anyscale
from anyscale._private.anyscale_client import FakeAnyscaleClient
from anyscale._private.sdk.timer import FakeTimer
from anyscale.client.openapi_client.models.session_ssh_key import SessionSshKey
from anyscale.compute_config.models import ComputeConfig, HeadNodeConfig
from anyscale.workspace import WorkspaceSDK
from anyscale.workspace.models import (
    UpdateWorkspaceConfig,
    WorkspaceConfig,
    WorkspaceState,
)


@pytest.fixture()
def workspace_sdk_with_fakes(
    sdk_with_fakes: Tuple[Anyscale, FakeAnyscaleClient, TestLogger, FakeTimer]
) -> Tuple[WorkspaceSDK, FakeAnyscaleClient, TestLogger, FakeTimer]:
    sdk, client, logger, timer = sdk_with_fakes
    return sdk.workspace, client, logger, timer


EXPECTED_SSH_TEMPLATE = """
Host Head-Node
  HostName {head_node_ip}
  User ubuntu
  IdentityFile {key_path}
  StrictHostKeyChecking false
  IdentitiesOnly yes

Host {name}
  HostName 0.0.0.0
  ProxyJump Head-Node
  Port 5020
  User ray
  IdentityFile {key_path}
  StrictHostKeyChecking false
  IdentitiesOnly yes
"""


class TestCreate:
    @pytest.mark.parametrize(
        "config", [WorkspaceConfig(name=None, project=None, cloud=None), None,],
    )
    def test_missing_arg(
        self,
        workspace_sdk_with_fakes: Tuple[
            WorkspaceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
        config: Optional[WorkspaceConfig],
    ):
        sdk, fake_client, _, _ = workspace_sdk_with_fakes

        with pytest.raises(ValueError, match="Workspace name must be configured"):
            sdk.create(config=config)

    def test_basic(
        self,
        workspace_sdk_with_fakes: Tuple[
            WorkspaceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = workspace_sdk_with_fakes

        workspace_id = sdk.create(config=WorkspaceConfig(name="test_workspace"))
        assert workspace_id is not None
        created_workspace = fake_client.get_workspace(id=workspace_id)
        assert created_workspace is not None
        assert created_workspace.name == "test_workspace"

    def test_with_container_image(
        self,
        workspace_sdk_with_fakes: Tuple[
            WorkspaceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = workspace_sdk_with_fakes

        workspace_id = sdk.create(
            config=WorkspaceConfig(
                name="test_workspace", containerfile=TEST_CONTAINERFILE,
            )
        )
        assert workspace_id is not None
        created_workspace = fake_client.get_workspace(id=workspace_id)

        builds = fake_client.get_non_default_cluster_env_builds()
        assert len(builds) == 1
        assert created_workspace is not None
        assert created_workspace.name == "test_workspace"
        assert created_workspace.environment_id == builds[0].id

    def test_with_image_uri(
        self,
        workspace_sdk_with_fakes: Tuple[
            WorkspaceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = workspace_sdk_with_fakes

        workspace_id = sdk.create(
            config=WorkspaceConfig(name="test_workspace", image_uri="test_image_uri",)
        )
        assert workspace_id is not None
        created_workspace = fake_client.get_workspace(id=workspace_id)
        builds = fake_client.get_non_default_cluster_env_builds()
        assert len(builds) == 1
        assert created_workspace is not None
        assert created_workspace.name == "test_workspace"
        assert created_workspace.environment_id == builds[0].id

    def test_with_compute_config(
        self,
        workspace_sdk_with_fakes: Tuple[
            WorkspaceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = workspace_sdk_with_fakes

        workspace_id = sdk.create(
            config=WorkspaceConfig(
                name="test_workspace",
                compute_config=ComputeConfig(
                    cloud=fake_client.DEFAULT_CLOUD_NAME,
                    head_node=HeadNodeConfig(
                        instance_type="head-node-instance-type", flags={},
                    ),
                    flags={},
                ),
            )
        )
        assert workspace_id is not None
        created_workspace = fake_client.get_workspace(id=workspace_id)
        workspace_compute_config_id = created_workspace.compute_config_id
        compute_config = fake_client.get_compute_config(workspace_compute_config_id)
        assert created_workspace is not None
        assert created_workspace.name == "test_workspace"
        assert compute_config.config.cloud_id == fake_client.DEFAULT_CLOUD_ID
        assert (
            compute_config.config.head_node_type.instance_type
            == "head-node-instance-type"
        )

    def test_with_idle_termination(
        self,
        workspace_sdk_with_fakes: Tuple[
            WorkspaceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = workspace_sdk_with_fakes

        workspace_id = sdk.create(
            config=WorkspaceConfig(name="test_workspace", idle_termination_minutes=180,)
        )
        assert workspace_id is not None
        created_workspace = fake_client.get_workspace(id=workspace_id)

        workspace_compute_config_id = created_workspace.compute_config_id
        compute_config = fake_client.get_compute_config(workspace_compute_config_id)
        assert created_workspace is not None
        assert created_workspace.name == "test_workspace"
        assert compute_config.idle_timeout_minutes == 180

    def test_with_requirements(
        self,
        workspace_sdk_with_fakes: Tuple[
            WorkspaceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = workspace_sdk_with_fakes

        workspace_id = sdk.create(
            config=WorkspaceConfig(name="test_workspace", requirements=["emoji"],)
        )
        assert workspace_id is not None
        created_workspace = fake_client.get_workspace(id=workspace_id)

        assert created_workspace is not None
        assert created_workspace.name == "test_workspace"
        assert fake_client._workspaces_dependencies[created_workspace.id] == ["emoji"]

    def test_with_env_vars(
        self,
        workspace_sdk_with_fakes: Tuple[
            WorkspaceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = workspace_sdk_with_fakes

        workspace_id = sdk.create(
            config=WorkspaceConfig(name="test_workspace", env_vars={"key": "value"},)
        )
        assert workspace_id is not None
        created_workspace = fake_client.get_workspace(id=workspace_id)

        assert created_workspace is not None
        assert created_workspace.name == "test_workspace"
        assert fake_client._workspaces_env_vars[created_workspace.id] == {
            "key": "value"
        }

    def test_with_full_config(
        self,
        workspace_sdk_with_fakes: Tuple[
            WorkspaceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = workspace_sdk_with_fakes

        workspace_id = sdk.create(
            config=WorkspaceConfig(
                name="test_workspace",
                idle_termination_minutes=180,
                containerfile=TEST_CONTAINERFILE,
                compute_config=ComputeConfig(
                    cloud=fake_client.DEFAULT_CLOUD_NAME,
                    head_node=HeadNodeConfig(
                        instance_type="head-node-instance-type", flags={},
                    ),
                    flags={},
                ),
                env_vars={"key": "value"},
                requirements=["emoji"],
            )
        )
        assert workspace_id is not None
        created_workspace = fake_client.get_workspace(id=workspace_id)

        builds = fake_client.get_non_default_cluster_env_builds()
        workspace_compute_config_id = created_workspace.compute_config_id
        compute_config = fake_client.get_compute_config(workspace_compute_config_id)
        assert len(builds) == 1
        assert created_workspace is not None
        assert created_workspace.name == "test_workspace"
        assert created_workspace.environment_id == builds[0].id
        assert created_workspace.cloud_id == fake_client.DEFAULT_CLOUD_ID
        assert created_workspace.creator_id == fake_client.DEFAULT_USER_ID
        assert created_workspace.creator_email == fake_client.DEFAULT_USER_EMAIL
        assert created_workspace.organization_id == fake_client.DEFAULT_ORGANIZATION_ID

        # verify compute config
        assert compute_config.idle_timeout_minutes == 180
        assert compute_config.config.cloud_id == fake_client.DEFAULT_CLOUD_ID
        assert (
            compute_config.config.head_node_type.instance_type
            == "head-node-instance-type"
        )

        # verify dependencies
        assert fake_client._workspaces_dependencies[created_workspace.id] == ["emoji"]
        assert fake_client._workspaces_env_vars[created_workspace.id] == {
            "key": "value"
        }


class TestStart:
    def test_missing_arg(
        self,
        workspace_sdk_with_fakes: Tuple[
            WorkspaceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = workspace_sdk_with_fakes

        with pytest.raises(ValueError, match="One of 'name' or 'id' must be provided."):
            sdk.start(name=None, id=None, cloud=None, project=None)

    def test_basic_with_id(
        self,
        workspace_sdk_with_fakes: Tuple[
            WorkspaceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = workspace_sdk_with_fakes

        # create first
        workspace_id = sdk.create(config=WorkspaceConfig(name="test_workspace"))
        assert workspace_id is not None

        sdk.start(id=workspace_id)
        created_workspace_status = sdk.status(id=workspace_id)
        assert created_workspace_status == WorkspaceState.RUNNING

    def test_basic_with_name(
        self,
        workspace_sdk_with_fakes: Tuple[
            WorkspaceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = workspace_sdk_with_fakes

        # create first
        workspace_id = sdk.create(config=WorkspaceConfig(name="test_workspace"))
        assert workspace_id is not None

        sdk.start(name="test_workspace")
        created_workspace_status = sdk.status(id=workspace_id)
        assert created_workspace_status == WorkspaceState.RUNNING


class TestTerminate:
    def test_missing_arg(
        self,
        workspace_sdk_with_fakes: Tuple[
            WorkspaceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = workspace_sdk_with_fakes

        with pytest.raises(ValueError, match="One of 'name' or 'id' must be provided."):
            sdk.terminate(name=None, id=None, cloud=None, project=None)

    def test_basic_with_id(
        self,
        workspace_sdk_with_fakes: Tuple[
            WorkspaceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = workspace_sdk_with_fakes

        # create first
        workspace_id = sdk.create(config=WorkspaceConfig(name="test_workspace"))
        assert workspace_id is not None

        sdk.terminate(id=workspace_id)
        created_workspace_status = sdk.status(id=workspace_id)
        assert created_workspace_status == WorkspaceState.TERMINATED

    def test_basic_with_name(
        self,
        workspace_sdk_with_fakes: Tuple[
            WorkspaceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = workspace_sdk_with_fakes

        # create first
        workspace_id = sdk.create(config=WorkspaceConfig(name="test_workspace"))
        assert workspace_id is not None

        sdk.terminate(name="test_workspace")
        created_workspace_status = sdk.status(id=workspace_id)
        assert created_workspace_status == WorkspaceState.TERMINATED


class TestStatus:
    def test_missing_arg(
        self,
        workspace_sdk_with_fakes: Tuple[
            WorkspaceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = workspace_sdk_with_fakes

        with pytest.raises(ValueError, match="One of 'name' or 'id' must be provided."):
            sdk.status(name=None, id=None, cloud=None, project=None)

    def test_basic_with_id(
        self,
        workspace_sdk_with_fakes: Tuple[
            WorkspaceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = workspace_sdk_with_fakes

        # create first
        workspace_id = sdk.create(config=WorkspaceConfig(name="test_workspace"))
        assert workspace_id is not None

        status = sdk.status(id=workspace_id)
        assert status == "TERMINATED"

    def test_basic_with_name(
        self,
        workspace_sdk_with_fakes: Tuple[
            WorkspaceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = workspace_sdk_with_fakes

        # create first
        workspace_id = sdk.create(config=WorkspaceConfig(name="test_workspace"))
        assert workspace_id is not None

        status = sdk.status(name="test_workspace")
        assert status == "TERMINATED"

    def test_not_found(
        self,
        workspace_sdk_with_fakes: Tuple[
            WorkspaceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = workspace_sdk_with_fakes

        name = "test_workspace"
        with pytest.raises(
            ValueError, match=f"Workspace with name '{name}' was not found."
        ):
            sdk.status(name=name)


class TestWait:
    def test_missing_arg(
        self,
        workspace_sdk_with_fakes: Tuple[
            WorkspaceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = workspace_sdk_with_fakes

        with pytest.raises(ValueError, match="One of 'name' or 'id' must be provided."):
            sdk.wait(name=None, id=None, cloud=None, project=None)

    def test_basic_with_id(
        self,
        workspace_sdk_with_fakes: Tuple[
            WorkspaceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = workspace_sdk_with_fakes

        # create first
        workspace_id = sdk.create(config=WorkspaceConfig(name="test_workspace"))
        assert workspace_id is not None

        # start
        sdk.start(id=workspace_id)

        sdk.wait(id=workspace_id)
        created_workspace_status = sdk.status(id=workspace_id)
        assert created_workspace_status == WorkspaceState.RUNNING

    def test_basic_with_name(
        self,
        workspace_sdk_with_fakes: Tuple[
            WorkspaceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = workspace_sdk_with_fakes

        # create first
        workspace_id = sdk.create(config=WorkspaceConfig(name="test_workspace"))
        assert workspace_id is not None

        # start
        sdk.start(id=workspace_id)

        sdk.wait(name="test_workspace")
        created_workspace_status = sdk.status(id=workspace_id)
        assert created_workspace_status == WorkspaceState.RUNNING

    def test_not_found(
        self,
        workspace_sdk_with_fakes: Tuple[
            WorkspaceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = workspace_sdk_with_fakes

        name = "test_workspace"
        with pytest.raises(
            ValueError, match=f"Workspace with name '{name}' was not found."
        ):
            sdk.wait(name=name)

    def test_custom_state(
        self,
        workspace_sdk_with_fakes: Tuple[
            WorkspaceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = workspace_sdk_with_fakes

        # create first
        workspace_id = sdk.create(config=WorkspaceConfig(name="test_workspace"))
        assert workspace_id is not None

        sdk.wait(id=workspace_id, state="TERMINATED")
        created_workspace_status = sdk.status(id=workspace_id)
        assert created_workspace_status == WorkspaceState.TERMINATED

    def test_timeout(
        self,
        workspace_sdk_with_fakes: Tuple[
            WorkspaceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, _, _, _ = workspace_sdk_with_fakes

        # create first
        workspace_id = sdk.create(config=WorkspaceConfig(name="test_workspace"))
        assert workspace_id is not None

        # wait on a workspace that is not started and verify a timeout error is raised
        with pytest.raises(
            TimeoutError,
            match=re.escape(
                f"Workspace '{workspace_id}' did not reach target state RUNNING within 1s. Last seen state: TERMINATED."
            ),
        ):
            sdk.wait(id=workspace_id, state="RUNNING", timeout_s=1)


class TestGenerateSSHConfigFile:
    def test_missing_arg(
        self,
        workspace_sdk_with_fakes: Tuple[
            WorkspaceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = workspace_sdk_with_fakes

        with pytest.raises(ValueError, match="One of 'name' or 'id' must be provided."):
            sdk.generate_ssh_config_file(name=None, id=None, cloud=None, project=None)

    def test_basic_with_id(
        self,
        workspace_sdk_with_fakes: Tuple[
            WorkspaceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = workspace_sdk_with_fakes

        # create first
        workspace_id = sdk.create(config=WorkspaceConfig(name="test_workspace"))
        assert workspace_id is not None

        # start it
        sdk.start(id=workspace_id)

        # update the fake client
        fake_workspace = fake_client._workspaces[workspace_id]
        fake_client._clusters_headnode_ip[fake_workspace.cluster_id] = "1.1.1.1"
        fake_client._clusters_ssh_key[fake_workspace.cluster_id] = SessionSshKey(
            key_name="fake_key_name", private_key="fake_key",
        )

        with tempfile.TemporaryDirectory() as temp_dir:
            host_name, config_path = sdk.generate_ssh_config_file(
                id=workspace_id, ssh_config_path=temp_dir
            )
            assert host_name == fake_workspace.name

            verify_file_content(
                config_path,
                EXPECTED_SSH_TEMPLATE.format(
                    name=fake_workspace.name,
                    head_node_ip="1.1.1.1",
                    key_path=os.path.join(temp_dir, "fake_key_name.pem",),
                ),
            )

    def test_basic_with_name(
        self,
        workspace_sdk_with_fakes: Tuple[
            WorkspaceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = workspace_sdk_with_fakes

        # create first
        workspace_id = sdk.create(config=WorkspaceConfig(name="test_workspace"))
        assert workspace_id is not None

        # start it
        sdk.start(id=workspace_id)

        # update the fake client
        fake_workspace = fake_client._workspaces[workspace_id]
        fake_client._clusters_headnode_ip[fake_workspace.cluster_id] = "1.1.1.1"
        fake_client._clusters_ssh_key[fake_workspace.cluster_id] = SessionSshKey(
            key_name="fake_key_name", private_key="fake_key",
        )

        with tempfile.TemporaryDirectory() as temp_dir:
            host_name, config_path = sdk.generate_ssh_config_file(
                id=workspace_id, ssh_config_path=temp_dir
            )
            assert host_name == fake_workspace.name

            verify_file_content(
                config_path,
                EXPECTED_SSH_TEMPLATE.format(
                    name=fake_workspace.name,
                    head_node_ip="1.1.1.1",
                    key_path=os.path.join(temp_dir, "fake_key_name.pem",),
                ),
            )

    def test_not_found(
        self,
        workspace_sdk_with_fakes: Tuple[
            WorkspaceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = workspace_sdk_with_fakes

        name = "test_workspace"
        with pytest.raises(
            ValueError, match=f"Workspace with name '{name}' was not found."
        ):
            sdk.generate_ssh_config_file(name=name)


class TestRunCommand:
    def test_missing_arg(
        self,
        workspace_sdk_with_fakes: Tuple[
            WorkspaceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = workspace_sdk_with_fakes

        with pytest.raises(ValueError, match="One of 'name' or 'id' must be provided."):
            sdk.run_command(
                name=None, id=None, cloud=None, project=None, command="ray --version"
            )

    def test_basic_with_id(
        self,
        workspace_sdk_with_fakes: Tuple[
            WorkspaceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = workspace_sdk_with_fakes

        # create first
        workspace_id = sdk.create(config=WorkspaceConfig(name="test_workspace"))
        assert workspace_id is not None

        # start it
        sdk.start(id=workspace_id)

        command = "ray --version"

        # update the fake client
        fake_workspace = fake_client._workspaces[workspace_id]
        fake_client._clusters_headnode_ip[fake_workspace.cluster_id] = "1.1.1.1"
        fake_client._clusters_ssh_key[fake_workspace.cluster_id] = SessionSshKey(
            key_name="fake_key_name", private_key="fake_key",
        )

        # mock subprocess.run and verify it is called
        with patch("subprocess.run") as mock_subprocess_run:
            sdk.run_command(id=workspace_id, command=command)
            expected_call_subset = ["ssh", "-F", mock.ANY, "test_workspace", command]
            first_call = mock_subprocess_run.call_args_list[0]
            contains_expected_call = all(
                arg in first_call[0][0] for arg in expected_call_subset
            )
            assert (
                contains_expected_call
            ), "The expected subset of arguments was not found in the actual calls."

    def test_basic_with_extra_run_args(
        self,
        workspace_sdk_with_fakes: Tuple[
            WorkspaceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = workspace_sdk_with_fakes

        # create first
        workspace_id = sdk.create(config=WorkspaceConfig(name="test_workspace"))
        assert workspace_id is not None

        # start it
        sdk.start(id=workspace_id)

        command = "ray --version"

        # update the fake client
        fake_workspace = fake_client._workspaces[workspace_id]
        fake_client._clusters_headnode_ip[fake_workspace.cluster_id] = "1.1.1.1"
        fake_client._clusters_ssh_key[fake_workspace.cluster_id] = SessionSshKey(
            key_name="fake_key_name", private_key="fake_key",
        )

        # mock subprocess.run and verify it is called
        with patch("subprocess.run") as mock_subprocess_run:
            sdk.run_command(
                id=workspace_id, command=command, check=False, capture_output=True
            )
            expected_call_subset = ["ssh", "-F", mock.ANY, "test_workspace", command]
            first_call = mock_subprocess_run.call_args_list[0]
            contains_expected_call = all(
                arg in first_call[0][0] for arg in expected_call_subset
            )
            assert (
                contains_expected_call
            ), "The expected subset of arguments was not found in the actual calls."
            assert first_call[1]["check"] is False
            assert first_call[1]["capture_output"] is True

    def test_not_found(
        self,
        workspace_sdk_with_fakes: Tuple[
            WorkspaceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = workspace_sdk_with_fakes

        name = "test_workspace"
        with pytest.raises(
            ValueError, match=f"Workspace with name '{name}' was not found."
        ):
            sdk.run_command(name=name, command="ray --version")


class TestPull:
    def test_missing_arg(
        self,
        workspace_sdk_with_fakes: Tuple[
            WorkspaceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = workspace_sdk_with_fakes

        with pytest.raises(ValueError, match="One of 'name' or 'id' must be provided."):
            sdk.pull(name=None, id=None, cloud=None, project=None)

    def test_basic_with_id(
        self,
        workspace_sdk_with_fakes: Tuple[
            WorkspaceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = workspace_sdk_with_fakes

        # create first
        workspace_id = sdk.create(config=WorkspaceConfig(name="test_workspace"))
        assert workspace_id is not None

        # start it
        sdk.start(id=workspace_id)

        # update the fake client
        fake_workspace = fake_client._workspaces[workspace_id]
        fake_client._clusters_headnode_ip[fake_workspace.cluster_id] = "1.1.1.1"
        fake_client._clusters_ssh_key[fake_workspace.cluster_id] = SessionSshKey(
            key_name="fake_key_name", private_key="fake_key",
        )

        # mock subprocess.run and verify rsync is called
        with patch("subprocess.run") as mock_subprocess_run:
            local_dir = "/path/to/local/dir"
            sdk.pull(id=workspace_id, local_dir=local_dir)

            expected_call_subset = [
                "rsync",
                "-rvzl",
                "-e",
                mock.ANY,
                "ray@test_workspace:~/default/",
                local_dir,
                "--delete",
                "--exclude",
                ".git",
                "--exclude",
                ".git/objects/info/alternates",
                "--exclude",
                ".anyscale.yaml",
            ]

            first_call = mock_subprocess_run.call_args_list[0]
            contains_expected_call = all(
                arg in first_call[0][0] for arg in expected_call_subset
            )
            assert (
                contains_expected_call
            ), "The expected subset of arguments was not found in the actual calls."

    def test_basic_with_pull_git_state(
        self,
        workspace_sdk_with_fakes: Tuple[
            WorkspaceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = workspace_sdk_with_fakes

        # create first
        workspace_id = sdk.create(config=WorkspaceConfig(name="test_workspace"))
        assert workspace_id is not None

        # start it
        sdk.start(id=workspace_id)

        # update the fake client
        fake_workspace = fake_client._workspaces[workspace_id]
        fake_client._clusters_headnode_ip[fake_workspace.cluster_id] = "1.1.1.1"
        fake_client._clusters_ssh_key[fake_workspace.cluster_id] = SessionSshKey(
            key_name="fake_key_name", private_key="fake_key",
        )

        # mock subprocess.run and verify rsync is called
        with patch("subprocess.run") as mock_subprocess_run:
            local_dir = "/path/to/local/dir"
            sdk.pull(id=workspace_id, local_dir=local_dir, pull_git_state=True)

            # verify repacing git first
            repac_call = mock_subprocess_run.call_args_list[1]
            expected_call_subset = [
                "python",
                "-m",
                "snapshot_util",
                "repack_git_repos",
            ]
            contains_expected_call = all(
                arg in repac_call[0][0] for arg in expected_call_subset
            )

            # verify rsync call
            expected_call_subset = [
                "rsync",
                "-rvzl",
                "-e",
                mock.ANY,
                "ray@test_workspace:~/default/",
                local_dir,
                "--delete",
            ]

            rsync_call = mock_subprocess_run.call_args_list[1]
            contains_expected_call = all(
                arg in rsync_call[0][0] for arg in expected_call_subset
            )
            assert (
                contains_expected_call
            ), "The expected subset of arguments was not found in the actual calls."

            # verify git is not excluded
            assert ".git" not in rsync_call[0][0]

    def test_basic_with_extra_rsync_args(
        self,
        workspace_sdk_with_fakes: Tuple[
            WorkspaceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = workspace_sdk_with_fakes

        # create first
        workspace_id = sdk.create(config=WorkspaceConfig(name="test_workspace"))
        assert workspace_id is not None

        # start it
        sdk.start(id=workspace_id)

        # update the fake client
        fake_workspace = fake_client._workspaces[workspace_id]
        fake_client._clusters_headnode_ip[fake_workspace.cluster_id] = "1.1.1.1"
        fake_client._clusters_ssh_key[fake_workspace.cluster_id] = SessionSshKey(
            key_name="fake_key_name", private_key="fake_key",
        )

        # mock subprocess.run and verify rsync is called
        with patch("subprocess.run") as mock_subprocess_run:
            local_dir = "/path/to/local/dir"
            rsync_args = ["--exclude", "file"]
            sdk.pull(id=workspace_id, local_dir=local_dir, rsync_args=rsync_args)

            expected_call_subset = [
                "rsync",
                "-rvzl",
                "-e",
                mock.ANY,
                "ray@test_workspace:~/default/",
                local_dir,
                "--exclude",
                "file",
            ]
            first_call = mock_subprocess_run.call_args_list[0]
            contains_expected_call = all(
                arg in first_call[0][0] for arg in expected_call_subset
            )
            assert (
                contains_expected_call
            ), "The expected subset of arguments was not found in the actual calls."


class TestPush:
    def test_missing_arg(
        self,
        workspace_sdk_with_fakes: Tuple[
            WorkspaceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = workspace_sdk_with_fakes

        with pytest.raises(ValueError, match="One of 'name' or 'id' must be provided."):
            sdk.push(name=None, id=None, cloud=None, project=None)

    def test_basic_with_id(
        self,
        workspace_sdk_with_fakes: Tuple[
            WorkspaceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = workspace_sdk_with_fakes

        # create first
        workspace_id = sdk.create(config=WorkspaceConfig(name="test_workspace"))
        assert workspace_id is not None

        # start it
        sdk.start(id=workspace_id)

        # update the fake client
        fake_workspace = fake_client._workspaces[workspace_id]
        fake_client._clusters_headnode_ip[fake_workspace.cluster_id] = "1.1.1.1"
        fake_client._clusters_ssh_key[fake_workspace.cluster_id] = SessionSshKey(
            key_name="fake_key_name", private_key="fake_key",
        )

        # mock subprocess.run and verify rsync is called
        with patch("subprocess.run") as mock_subprocess_run:
            local_dir = "/path/to/local/dir/"
            sdk.push(id=workspace_id, local_dir=local_dir)

            expected_call_subset = [
                "rsync",
                "-rvzl",
                "-e",
                mock.ANY,
                local_dir,
                "ray@test_workspace:~/default/",
                "--delete",
                "--exclude",
                ".git",
                "--exclude",
                ".git/objects/info/alternates",
                "--exclude",
                ".anyscale.yaml",
            ]

            first_call = mock_subprocess_run.call_args_list[0]
            contains_expected_call = all(
                arg in first_call[0][0] for arg in expected_call_subset
            )
            assert (
                contains_expected_call
            ), "The expected subset of arguments was not found in the actual calls."

    def test_basic_with_push_git_state(
        self,
        workspace_sdk_with_fakes: Tuple[
            WorkspaceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = workspace_sdk_with_fakes

        # create first
        workspace_id = sdk.create(config=WorkspaceConfig(name="test_workspace"))
        assert workspace_id is not None

        # start it
        sdk.start(id=workspace_id)

        # update the fake client
        fake_workspace = fake_client._workspaces[workspace_id]
        fake_client._clusters_headnode_ip[fake_workspace.cluster_id] = "1.1.1."
        fake_client._clusters_ssh_key[fake_workspace.cluster_id] = SessionSshKey(
            key_name="fake_key_name", private_key="fake_key",
        )

        # mock subprocess.run and verify rsync is called
        with patch("subprocess.run") as mock_subprocess_run:
            local_dir = "/path/to/local/dir/"
            sdk.push(id=workspace_id, local_dir=local_dir, push_git_state=True)

            expected_call_subset = [
                "rsync",
                "-rvzl",
                "-e",
                mock.ANY,
                local_dir,
                "ray@test_workspace:~/default/",
                "--delete",
                "--exclude",
                ".git/objects/info/alternates",
                "--exclude",
                ".anyscale.yaml",
            ]

            first_call = mock_subprocess_run.call_args_list[0]
            contains_expected_call = all(
                arg in first_call[0][0] for arg in expected_call_subset
            )
            assert (
                contains_expected_call
            ), "The expected subset of arguments was not found in the actual calls."

            # verify that the .git directory is not excluded
            assert ".git" not in first_call[0][0]

    def test_basic_with_extra_rsync_args(
        self,
        workspace_sdk_with_fakes: Tuple[
            WorkspaceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = workspace_sdk_with_fakes

        # create first
        workspace_id = sdk.create(config=WorkspaceConfig(name="test_workspace"))
        assert workspace_id is not None

        # start it
        sdk.start(id=workspace_id)

        # update the fake client
        fake_workspace = fake_client._workspaces[workspace_id]
        fake_client._clusters_headnode_ip[fake_workspace.cluster_id] = "1.1.1.1"
        fake_client._clusters_ssh_key[fake_workspace.cluster_id] = SessionSshKey(
            key_name="fake_key_name", private_key="fake_key",
        )

        # mock subprocess.run and verify rsync is called
        with patch("subprocess.run") as mock_subprocess_run:
            local_dir = "/path/to/local/dir/"
            rsync_args = ["--exclude", "file"]
            sdk.push(id=workspace_id, local_dir=local_dir, rsync_args=rsync_args)

            expected_call_subset = [
                "rsync",
                "-rvzl",
                "-e",
                mock.ANY,
                local_dir,
                "ray@test_workspace:~/default/",
                "--exclude",
                "file",
            ]
            first_call = mock_subprocess_run.call_args_list[0]
            contains_expected_call = all(
                arg in first_call[0][0] for arg in expected_call_subset
            )
            assert (
                contains_expected_call
            ), "The expected subset of arguments was not found in the actual calls."


class TestUpdate:
    def test_basic(
        self,
        workspace_sdk_with_fakes: Tuple[
            WorkspaceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = workspace_sdk_with_fakes

        # create a workspace
        workspace_id = sdk.create(config=WorkspaceConfig(name="test_workspace"))
        assert workspace_id is not None

        sdk.update(
            id=workspace_id, config=UpdateWorkspaceConfig(name="test_workspace_2")
        )
        created_workspace = fake_client.get_workspace(id=workspace_id)
        assert created_workspace is not None
        assert created_workspace.name == "test_workspace_2"

    def test_with_container_image(
        self,
        workspace_sdk_with_fakes: Tuple[
            WorkspaceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = workspace_sdk_with_fakes

        workspace_id = sdk.create(config=WorkspaceConfig(name="test_workspace"))
        assert workspace_id is not None

        sdk.update(
            id=workspace_id,
            config=UpdateWorkspaceConfig(containerfile=TEST_CONTAINERFILE),
        )

        created_workspace = fake_client.get_workspace(id=workspace_id)

        builds = fake_client.get_non_default_cluster_env_builds()
        assert len(builds) == 1
        assert created_workspace is not None
        assert created_workspace.name == "test_workspace"
        assert created_workspace.environment_id == builds[0].id

    def test_with_image_uri(
        self,
        workspace_sdk_with_fakes: Tuple[
            WorkspaceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = workspace_sdk_with_fakes

        workspace_id = sdk.create(config=WorkspaceConfig(name="test_workspace"))
        assert workspace_id is not None

        sdk.update(
            id=workspace_id, config=UpdateWorkspaceConfig(image_uri="test_image_uri",)
        )

        updated_workspace = fake_client.get_workspace(id=workspace_id)
        builds = fake_client.get_non_default_cluster_env_builds()
        assert len(builds) == 1
        assert updated_workspace is not None
        assert updated_workspace.name == "test_workspace"
        assert updated_workspace.environment_id == builds[0].id

    def test_with_compute_config(
        self,
        workspace_sdk_with_fakes: Tuple[
            WorkspaceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = workspace_sdk_with_fakes

        workspace_id = sdk.create(config=WorkspaceConfig(name="test_workspace",))
        assert workspace_id is not None

        sdk.update(
            id=workspace_id,
            config=UpdateWorkspaceConfig(
                compute_config=ComputeConfig(
                    cloud=fake_client.DEFAULT_CLOUD_NAME,
                    head_node=HeadNodeConfig(
                        instance_type="head-node-instance-type", flags={},
                    ),
                    flags={},
                ),
            ),
        )

        updated_workspace = fake_client.get_workspace(id=workspace_id)
        workspace_compute_config_id = updated_workspace.compute_config_id
        compute_config = fake_client.get_compute_config(workspace_compute_config_id)
        assert updated_workspace is not None
        assert updated_workspace.name == "test_workspace"
        assert compute_config.config.cloud_id == fake_client.DEFAULT_CLOUD_ID
        assert (
            compute_config.config.head_node_type.instance_type
            == "head-node-instance-type"
        )

    def test_with_idle_termination(
        self,
        workspace_sdk_with_fakes: Tuple[
            WorkspaceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = workspace_sdk_with_fakes

        workspace_id = sdk.create(
            config=WorkspaceConfig(name="test_workspace", idle_termination_minutes=180,)
        )
        assert workspace_id is not None

        sdk.update(
            id=workspace_id, config=UpdateWorkspaceConfig(idle_termination_minutes=500,)
        )
        updated_workspace = fake_client.get_workspace(id=workspace_id)

        workspace_compute_config_id = updated_workspace.compute_config_id
        compute_config = fake_client.get_compute_config(workspace_compute_config_id)
        assert updated_workspace is not None
        assert updated_workspace.name == "test_workspace"
        assert compute_config.idle_timeout_minutes == 500

    def test_with_requirements(
        self,
        workspace_sdk_with_fakes: Tuple[
            WorkspaceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = workspace_sdk_with_fakes

        workspace_id = sdk.create(
            config=WorkspaceConfig(name="test_workspace", requirements=["emoji"],)
        )
        assert workspace_id is not None

        sdk.update(
            id=workspace_id, config=UpdateWorkspaceConfig(requirements=["pytorch"],)
        )
        updated_workspace = fake_client.get_workspace(id=workspace_id)

        assert updated_workspace is not None
        assert updated_workspace.name == "test_workspace"
        assert fake_client._workspaces_dependencies[updated_workspace.id] == ["pytorch"]

    def test_with_env_vars(
        self,
        workspace_sdk_with_fakes: Tuple[
            WorkspaceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = workspace_sdk_with_fakes

        workspace_id = sdk.create(
            config=WorkspaceConfig(name="test_workspace", env_vars={"key": "value"},)
        )
        assert workspace_id is not None

        sdk.update(
            id=workspace_id,
            config=UpdateWorkspaceConfig(
                name="test_workspace", env_vars={"key2": "value2"},
            ),
        )
        updated_workspace = fake_client.get_workspace(id=workspace_id)

        assert updated_workspace is not None
        assert updated_workspace.name == "test_workspace"
        assert fake_client._workspaces_env_vars[updated_workspace.id] == {
            "key2": "value2"
        }

    def test_with_full_config(
        self,
        workspace_sdk_with_fakes: Tuple[
            WorkspaceSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = workspace_sdk_with_fakes

        workspace_id = sdk.create(config=WorkspaceConfig(name="test_workspace"))
        assert workspace_id is not None

        sdk.update(
            id=workspace_id,
            config=UpdateWorkspaceConfig(
                name="test_workspace_2",
                idle_termination_minutes=180,
                containerfile=TEST_CONTAINERFILE,
                compute_config=ComputeConfig(
                    cloud=fake_client.DEFAULT_CLOUD_NAME,
                    head_node=HeadNodeConfig(
                        instance_type="head-node-instance-type", flags={},
                    ),
                    flags={},
                ),
                env_vars={"key": "value"},
                requirements=["emoji"],
            ),
        )

        updated_workspace = fake_client.get_workspace(id=workspace_id)

        builds = fake_client.get_non_default_cluster_env_builds()
        workspace_compute_config_id = updated_workspace.compute_config_id
        compute_config = fake_client.get_compute_config(workspace_compute_config_id)
        assert len(builds) == 1
        assert updated_workspace is not None
        assert updated_workspace.name == "test_workspace_2"
        assert updated_workspace.environment_id == builds[0].id
        assert updated_workspace.cloud_id == fake_client.DEFAULT_CLOUD_ID
        assert updated_workspace.creator_id == fake_client.DEFAULT_USER_ID
        assert updated_workspace.creator_email == fake_client.DEFAULT_USER_EMAIL
        assert updated_workspace.organization_id == fake_client.DEFAULT_ORGANIZATION_ID

        # verify compute config
        assert compute_config.idle_timeout_minutes == 180
        assert compute_config.config.cloud_id == fake_client.DEFAULT_CLOUD_ID
        assert (
            compute_config.config.head_node_type.instance_type
            == "head-node-instance-type"
        )

        # verify dependencies
        assert fake_client._workspaces_dependencies[updated_workspace.id] == ["emoji"]
        assert fake_client._workspaces_env_vars[updated_workspace.id] == {
            "key": "value"
        }



from unittest.mock import patch

from asynctest import CoroutineMock
import pytest

from anyscale.api_utils import logs_util
from anyscale.api_utils.logs_util import (
    _download_log_from_s3_url,
    _remove_ansi_escape_sequences,
)


# Indirectly tests gather_in_batches()
async def test_download_logs_concurrently():
    # Test all downloads succeed
    with patch.object(
        logs_util,
        _download_log_from_s3_url.__name__,
        CoroutineMock(side_effect=["logs_from_chunk_1\n", "logs_from_chunk_2"]),
    ) as mock_download_log_from_s3_url:
        result = await logs_util._download_logs_concurrently(
            ["http://presigned.url.1", "http://presigned.url.2"], 1
        )
    assert result == "logs_from_chunk_1\nlogs_from_chunk_2"
    assert mock_download_log_from_s3_url.call_count == 2

    # Test one download fail raise exception
    exception_msg = "Log chunk download failed"
    with patch.object(
        logs_util,
        _download_log_from_s3_url.__name__,
        CoroutineMock(side_effect=Exception(exception_msg)),
    ) as mock_download_log_from_s3_url, pytest.raises(Exception, match=exception_msg):
        result = await logs_util._download_logs_concurrently(
            ["http://presigned.url.1", "http://presigned.url.2"], 1
        )


def test_remove_ansi_escape_sequences():
    string_with_ansi_escape_sequences = "\x1b[1m\x1b[31mHello World\x1B[0m\x1B[0m"
    assert (
        _remove_ansi_escape_sequences(string_with_ansi_escape_sequences)
        == "Hello World"
    )



from datetime import datetime
from unittest.mock import Mock, patch

import pytest

from anyscale.client.openapi_client import Project
from anyscale.client.openapi_client.models.cloud_providers import CloudProviders
from anyscale.client.openapi_client.models.create_resource_quota import (
    CreateResourceQuota,
)
from anyscale.client.openapi_client.models.list_resource_quotas_query import (
    ListResourceQuotasQuery,
)
from anyscale.client.openapi_client.models.mini_cloud import MiniCloud
from anyscale.client.openapi_client.models.mini_user import MiniUser
from anyscale.client.openapi_client.models.organization_collaborator import (
    OrganizationCollaborator,
)
from anyscale.client.openapi_client.models.organization_permission_level import (
    OrganizationPermissionLevel,
)
from anyscale.client.openapi_client.models.quota import Quota
from anyscale.client.openapi_client.models.resource_quota import ResourceQuota
from anyscale.client.openapi_client.models.resource_quota_status import (
    ResourceQuotaStatus,
)
from anyscale.controllers.resource_quota_controller import ResourceQuotaController
from anyscale.sdk.anyscale_client.models.page_query import PageQuery
from anyscale.sdk.anyscale_client.models.text_query import TextQuery
from tests.controllers.test_service_account_controller import (
    ListResponse,
    ListResponseMetadata,
)


test_organization_collaborator = OrganizationCollaborator(
    id="test_identity_id",
    name="name",
    permission_level=OrganizationPermissionLevel.COLLABORATOR,
    created_at=datetime(2024, 7, 17),
    email="test@anyscale.com",
    user_id="user_id",
)

test_resource_quota = ResourceQuota(
    id="resource_quota_id",
    name="quota_name",
    cloud_id="cloud_id",
    project_id="project_id",
    user_id="user_id",
    is_enabled=True,
    created_at=datetime(2024, 7, 17),
    deleted_at=None,
    quota=Quota(
        num_cpus=10,
        num_instances=20,
        num_gpus=30,
        num_accelerators={"L4": 5, "T4": 10},
    ),
    creator=MiniUser(
        id="user_id", email="test@anyscale..com", name="name", username="username",
    ),
    cloud=MiniCloud(id="cloud_id", name="cloud_name", provider=CloudProviders.AWS,),
)


@pytest.fixture()
def mock_api_client(project_test_data: Project) -> Mock:
    mock_api_client = Mock()

    mock_api_client.list_organization_collaborators_api_v2_organization_collaborators_get.return_value = ListResponse(
        results=[test_organization_collaborator],
        metadata=ListResponseMetadata(total=1),
    )
    mock_api_client.find_cloud_by_name_api_v2_clouds_find_by_name_post.return_value = Mock(
        result=Mock(id="cloud_id", name="cloud_name")
    )
    mock_api_client.find_project_by_project_name_api_v2_projects_find_by_name_get.return_value = ListResponse(
        results=[project_test_data], metadata=ListResponseMetadata(total=1)
    )
    mock_api_client.create_resource_quota_api_v2_resource_quotas_post.return_value = Mock(
        result=test_resource_quota
    )
    mock_api_client.search_resource_quotas_api_v2_resource_quotas_search_post.return_value = ListResponse(
        results=[test_resource_quota], metadata=ListResponseMetadata(total=1)
    )

    return mock_api_client


@pytest.fixture()
def mock_auth_api_client(mock_api_client: Mock, base_mock_anyscale_api_client: Mock):
    mock_auth_api_client = Mock(
        api_client=mock_api_client, anyscale_api_client=base_mock_anyscale_api_client,
    )
    with patch.multiple(
        "anyscale.controllers.base_controller",
        get_auth_api_client=Mock(return_value=mock_auth_api_client),
    ):
        yield


def test_create_resource_quota(mock_auth_api_client) -> None:
    resource_quota_controller = ResourceQuotaController()

    resource_quota_controller.create(
        "quota_name",
        "cloud_name",
        "project_name",
        "test@anyscale.com",
        10,
        20,
        30,
        {"L4": 5, "T4": 10},
    )

    resource_quota_controller.api_client.create_resource_quota_api_v2_resource_quotas_post.assert_called_once_with(
        CreateResourceQuota(
            name="quota_name",
            cloud_id="cloud_id",
            project_id="project_id",
            user_id="user_id",
            quota=Quota(
                num_cpus=10,
                num_instances=20,
                num_gpus=30,
                num_accelerators={"L4": 5, "T4": 10},
            ),
        )
    )


@pytest.mark.parametrize("cloud_name", ["cloud_name", None])
def test_list_resource_quotas(cloud_name, mock_auth_api_client) -> None:
    resource_quota_controller = ResourceQuotaController()

    resource_quota_controller.list_resource_quotas(
        name="quota_name", cloud=cloud_name, creator_id="user_id", is_enabled=True,
    )

    resource_quota_controller.api_client.search_resource_quotas_api_v2_resource_quotas_search_post.assert_called_once_with(
        ListResourceQuotasQuery(
            name=TextQuery("quota_name"),
            cloud_id="cloud_id" if cloud_name else None,
            creator_id="user_id",
            is_enabled=True,
            paging=PageQuery(count=20),
        )
    )


def test_delete(mock_auth_api_client) -> None:
    resource_quota_controller = ResourceQuotaController()
    mock_resource_quota_id = "resource_quota_id"

    resource_quota_controller.delete(mock_resource_quota_id)

    resource_quota_controller.api_client.delete_resource_quota_api_v2_resource_quotas_resource_quota_id_delete.assert_called_once_with(
        mock_resource_quota_id
    )


@pytest.mark.parametrize("is_enabled", [True, False])
def test_set_status(mock_auth_api_client, is_enabled: bool) -> None:
    resource_quota_controller = ResourceQuotaController()
    mock_resource_quota_id = "resource_quota_id"

    resource_quota_controller.set_status(mock_resource_quota_id, is_enabled=is_enabled)

    resource_quota_controller.api_client.set_resource_quota_status_api_v2_resource_quotas_resource_quota_id_status_patch.assert_called_once_with(
        mock_resource_quota_id, ResourceQuotaStatus(is_enabled=is_enabled)
    )
